import functools
import math
import numpy as np
from astropy.stats import sigma_clipped_stats


class FastExtraction:

    def __init__(self, image_layers, sigma=3, clipping=10):
        """\
        Parameters
        ----------
        image_layers: array_like of shape (num_layers, rows, columns)
        sigma: number
            Used for sigma clipping of the image background
        clipping: number
            After sigma clipping the layers are cut at mean + clipping * stddev
        """

        assert image_layers is not None and np.ndim(image_layers) == 3
        assert sigma > 0
        assert clipping > 0

        self.layers = np.asarray(image_layers)
        self.sigma = sigma
        self.clipping = clipping

    @property
    @functools.lru_cache(maxsize=None)
    def clipped_layers(self):
        return self._clip_image(self.layers, self.background_mean + self.background_stddev * self.clipping)

    @property
    @functools.lru_cache(maxsize=None)
    def _background_stats(self):
        return sigma_clipped_stats(self.layers, sigma_upper=self.sigma, sigma_lower=1000, cenfunc='mean', axis=(1, 2))

    @property
    def background_mean(self):
        return self._background_stats[0]

    @property
    def background_median(self):
        return self._background_stats[1]

    @property
    def background_stddev(self):
        return self._background_stats[2]

    @classmethod
    def _clip_image(cls, image, threshold):

        clipped = np.copy(image)
        clipped[np.isinf(clipped)] = np.nan

        n, __, __ = np.shape(image)

        threshold = np.reshape(threshold, (n, 1, 1))
        clipped[clipped < threshold] = np.nan

        return clipped

    @property
    def de_rotation_angles_deg(self):
        return np.rad2deg(self.de_rotation_angles_rad)

    @property
    @functools.lru_cache(maxsize=None)
    def de_rotation_angles_rad(self):
        return np.array([self._calculate_de_rotation_angle(layer) for layer in self.clipped_layers])

    @property
    @functools.lru_cache(maxsize=None)
    def de_rotated_layers(self):
        from scipy.ndimage.interpolation import rotate

        angle_deg = np.mean(self.de_rotation_angles_deg)
        return rotate(self.layers, angle_deg, axes=(1, 2), mode='constant', cval=np.nan)

    @property
    @functools.lru_cache(maxsize=None)
    def clipped_de_rotated_layers(self):
        mean, median, stddev = self._background_stats
        return self._clip_image(self.de_rotated_layers, mean + stddev * self.clipping)

    @classmethod
    def _calculate_de_rotation_angle(cls, image):
        """\
        Calculate image orientation using image moments as described in
        <https://en.wikipedia.org/wiki/Image_moment#Examples_2>.

        """
        assert image.ndim == 2

        indices_y, indices_x = np.indices(image.shape)

        m00 = np.nansum(image)
        m10 = np.nansum(image * indices_x)
        m01 = np.nansum(image * indices_y)
        m11 = np.nansum(image * indices_x * indices_y)
        m20 = np.nansum(image * indices_x * indices_x)
        m02 = np.nansum(image * indices_y * indices_y)

        avg_x = m10 / m00
        avg_y = m01 / m00
        mu_11_ = m11 / m00 - avg_x * avg_y
        mu_20_ = m20 / m00 - avg_x ** 2
        mu_02_ = m02 / m00 - avg_y ** 2

        angle = 0.5 * np.arctan2(2 * mu_11_, mu_20_ - mu_02_)

        while angle > math.pi / 2:
            angle -= math.pi

        while angle < -math.pi / 2:
            angle += math.pi

        return angle


def find_slit_in_images(rgb, background_mean, scale=1.0):

    __, size_y, __ = rgb.shape
    wo_background = rgb - np.reshape(background_mean, (-1, 1, 1))

    slit_function = np.nanmean(wo_background, axis=(0, 2))
    slit_center, slit_size = _center_of_gravity(slit_function)

    miny = math.floor(slit_center - scale * slit_size)
    maxy = math.ceil(slit_center + scale * slit_size)

    miny = np.clip(miny, 0, size_y - 1)
    maxy = np.clip(maxy, 0, size_y - 1)

    return miny, maxy


def find_spectra_in_layers(rgb, background_mean, scale=1.5):

    __, __, size_x = rgb.shape
    wo_background = rgb - np.reshape(background_mean, (-1, 1, 1))

    spectra = np.nanmean(wo_background, axis=1)

    spectrum_locations, spectrum_widths = _center_of_gravity(spectra)

    left_most_spectrum = np.argmin(spectrum_locations)
    right_most_spectrum = np.argmax(spectrum_locations)

    minx = math.floor(spectrum_locations[left_most_spectrum] - scale * spectrum_widths[left_most_spectrum])
    maxx = math.ceil(spectrum_locations[right_most_spectrum] + scale * spectrum_widths[right_most_spectrum])

    minx = np.clip(minx, 0, size_x - 1)
    maxx = np.clip(maxx, 0, size_x - 1)

    return minx, maxx


def _center_of_gravity(data):
    """\
    Return the weighted mean and standard deviation using moments.

    Mean is the first raw moment while variance is the second central moment of the data.
    For more details have a look at https://en.wikipedia.org/wiki/Moment_(mathematics) or
    https://en.wikipedia.org/wiki/Image_moment.
    .
    """

    if data.ndim > 1:
        result = [_center_of_gravity(data[i]) for i in range(data.shape[0])]
        result = np.transpose(result)
        return result[0], result[1]

    assert data.ndim == 1
    valid_indices = np.isfinite(data)

    indices = np.arange(data.shape[-1])
    indices = indices[valid_indices]
    data = data[valid_indices]

    mean = np.nansum(indices * data) / np.nansum(data)
    variance = np.nansum((indices - mean) ** 2 * data) / np.nansum(data)

    if variance < 0:
        # this may happen for negative data?
        variance = -variance

    return mean, math.sqrt(variance)



"""Display histogram of a raw image."""

import logging
import os.path
from argparse import ArgumentParser

import matplotlib.pyplot as plt
import numpy as np
from astropy.stats import sigma_clipped_stats

from bayer.to_rgb import rawpy_to_rgb
from bayer.utils import multi_glob


def main_raw():
    import rawpy

    parser = _create_argument_parser('one or more raw files containing bayer matrices')
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    for filename in multi_glob(args.filename):

        with rawpy.imread(filename) as raw:
            rgb = rawpy_to_rgb(raw)
            max_range = raw.white_level

        _plot_histogram(os.path.basename(filename), rgb, max_range, args.sigma, args.clipping)


def main_fits():
    from astropy.io import fits

    parser = _create_argument_parser('one or more fits files containing images')

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    for filename in multi_glob(args.filename):
        with fits.open(filename) as hdu_list:
            images = [hdu.data for hdu in hdu_list if hdu.header.get("NAXIS", 0) == 2]
            if not images:
                logging.error(f"{filename} contains no images")

            for image in images:
                _plot_histogram(os.path.basename(filename), np.asarray([image]), 2**16, args.sigma, args.clipping)


def _create_argument_parser(filename_help=None):
    parser = ArgumentParser(description='Display histogram of a raw image')
    parser.add_argument('filename', nargs='+', help=filename_help)
    parser.add_argument('--sigma', '-s', default=3.0, help='sigma used for clipping')
    parser.add_argument('--clipping', '-c', default=10.0, help='clip background at mean + clipping * stddev')
    return parser


def _plot_histogram(title, layers, max_range, sigma, clipping):

    fig = plt.figure()
    fig.canvas.manager.set_window_title(title)

    num_layers, __, __ = layers.shape

    ax = fig.add_subplot()
    if num_layers == 3:
        colors = 'rgb'
    else:
        colors = 'k' * num_layers

    for color, layer in zip(colors, [layers[i] for i in range(num_layers)]):
        (mean, median, stddev) = sigma_clipped_stats(layer, sigma=sigma)
        hist = np.histogram(layer[layer >= (mean + clipping * stddev)], bins=100)

        ax.plot(hist[1][1:], hist[0], color)

    ax.set_xlabel('intensity')
    ax.set_ylabel('pixel count')
    ax.set_yscale('log')
    ax.axvline(x=max_range, color='k', linestyle='--')
    ax.axvline(x=0.75 * max_range, color='k', linestyle='-.')
    fig.tight_layout(pad=0.5, h_pad=0.2, w_pad=0.2)
    plt.show()
    plt.close(fig)



import logging

import numpy as np

logger = logging.getLogger(__name__)


def rawpy_to_rgb(raw):
    """Extract RGB image from a rawpy bayer image."""

    assert all((c in raw.color_desc for c in b'RGB')), 'not a RBG raw image'

    layers = bayer_to_layers(raw.raw_image_visible, raw.raw_pattern)
    return combine_layers_by_color(layers, raw.color_desc, b'RGB')


def combine_layers_by_color(layers, layer_color_desc, target_color_desc=b'RGB', method='mean'):
    """\
    Fold layers by colors.

    Parameters
    ----------
    layers : array_like of shape (N, R, C)
        input image layers

    layer_color_desc : array_like of length N
        A single color for each layer, e.g. 'RGBG'

    target_color_desc: array_like
        A single color for each target layers, e.g. 'RGB'.
        It must only contain elements of layer_color_desc.

    method: str
        'mean', 'median' or any others numpy method of signature method(array, axis=...)
        It is used to combine source layers having the same color, e.g. the two green layers in a RGBG image.

    Return
    ------
    array_like with the same length as target_color_desc
    """
    assert len(layers) == len(layer_color_desc), f'length mismatch between layers and layer_color_desc'

    combiner = getattr(np, method, None)
    assert callable(combiner), f'np.{method} does not exist or is not callable'

    def combine_layers_of_color(color):
        layers_of_correct_color = layers[np.nonzero(np.array(list(layer_color_desc)) == color)]
        return combiner(layers_of_correct_color, axis=0)

    target = [combine_layers_of_color(color) for color in target_color_desc]
    return np.array(target)


def bayer_to_layers(bayer, pattern):
    """\
    Extract color layers from a raw bayer image an a pattern definition.

    If pattern is taken from rawpy.raw_pattern the resulting layers will match rawpy.color_desc.
    E.g. for a Canon D1000 color_desc='RGBG' while raw_pattern=[[0 1],[3 2]]. This results in
    four layers where each layer represents each second pixel, the first red-layer starts at (0,0), the second
    green-layer starts at (0,1) the third blue-layer starts at (1,1) and the forth layer, also green, starts at (1,0).


    Parameters
    ----------

    bayer : array_like of shape (r, c)
        The original raw bayer image

    pattern: array_like
        The smallest possible Bayer pattern of the image.

    """

    number_of_layers = np.max(pattern) + 1
    assert 0 <= np.min(pattern)

    layers = number_of_layers * [None]

    row_step_size, column_step_size = pattern.shape

    indices_y, indices_x = np.indices(pattern.shape)
    for start_row, start_column in zip(indices_y.ravel(), indices_x.ravel()):
        idx = pattern[start_row, start_column]
        layers[idx] = bayer[start_row::row_step_size, start_column::column_step_size]

    return np.asarray(layers)



"""\
Use image center-of-mass to extract spectra fast-and-dirty
"""

import logging
import os.path
from argparse import ArgumentParser

import matplotlib.pyplot as plt
import numpy as np

from bayer.extraction import FastExtraction
from bayer.to_rgb import rawpy_to_rgb
from bayer.utils import multi_glob
from bayer.extraction import find_slit_in_images, find_spectra_in_layers


def main_raw():
    import rawpy

    parser = _create_argument_parser('one or more raw file containing bayer matrices')

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    for filename in multi_glob(args.filename):
        with rawpy.imread(filename) as raw:
            extractor = FastExtraction(image_layers=rawpy_to_rgb(raw), sigma=args.sigma)
            _plot_file(filename, extractor, raw.white_level, args.cut)


def main_fits():
    from astropy.io import fits

    parser = _create_argument_parser('one or more fits files containing images')

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    for filename in multi_glob(args.filename):
        with fits.open(filename) as hdu_list:
            images = [hdu.data for hdu in hdu_list if hdu.header.get("NAXIS", 0) == 2]
            if not images:
                logging.error(f"{filename} contains no images")

            for image in images:
                extractor = FastExtraction(image_layers=[image], sigma=args.sigma)
                _plot_file(filename, extractor, 2**16, args.cut)


def _create_argument_parser(filename_help):
    parser = ArgumentParser(description='Display spectrum from a bayer matrix')
    parser.add_argument('filename', nargs='+', help=filename_help)
    parser.add_argument('--sigma', '-s', default=3.0, help='sigma used for clipping')
    parser.add_argument('--clipping', default=10.0, help='clip background at mean + clipping * stddev')
    cut_group = parser.add_mutually_exclusive_group()
    cut_group.add_argument('--cut', '-c', dest='cut', default=True, action='store_true',
                           help='cut spectrum in dispersion direction (the default)')
    cut_group.add_argument('--dont-cut', '-C', dest='cut', action='store_false',
                           help='do not cut spectrum in dispersion direction')
    return parser


def _plot_file(filename, extractor, white_level, cut_spectra):
    rgb = extractor.de_rotated_layers
    num_colors, size_y, size_x = rgb.shape

    miny, maxy = find_slit_in_images(rgb, extractor.background_mean)

    if cut_spectra:
        minx, maxx = find_spectra_in_layers(rgb, extractor.background_mean)
    else:
        minx, maxx = 0, size_x - 1

    rgb = rgb[:, miny:maxy, minx:maxx]
    (num_colors, size_y, size_x) = rgb.shape

    xrange = (0, size_x)

    fig = plt.figure()
    fig.canvas.manager.set_window_title(os.path.basename(filename))

    ax = plt.subplot2grid((12, 8), (0, 0), rowspan=10, colspan=7)
    ax.set_xlim(xrange)
    ax.get_xaxis().set_visible(False)

    if num_colors == 3:
        layers = rgb[0], rgb[1], rgb[2], rgb[0] + rgb[1] + rgb[2]
        colors = 'rgbk'
    else:
        layers = rgb
        colors = 'k' * num_colors

    for color, layer in zip(colors, layers):
        spec = np.nanmax(layer, axis=0)
        ax.plot(range(*spec.shape), spec, color)

    ax.axhline(y=white_level, color='k', linestyle='--')
    ax.axhline(y=0.75 * white_level, color='k', linestyle='-.')

    ax = plt.subplot2grid((12, 8), (10, 7))
    slit = np.nanmax(rgb, axis=(0, 2))
    ax.plot(slit, range(*slit.shape), 'k')
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)

    ax = plt.subplot2grid((12, 8), (10, 0), colspan=7)
    ax.imshow(_reshape_and_scale_image(rgb, white_level, scale=False), aspect='auto')
    ax.set_xlim(xrange)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)

    ax = plt.subplot2grid((12, 8), (11, 0), colspan=7)
    ax.imshow(_reshape_and_scale_image(rgb, white_level, scale=True), aspect='auto')
    ax.set_xlim(xrange)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)

    fig.tight_layout(pad=0.5, h_pad=0.2, w_pad=0.2)

    plt.show()
    plt.close(fig)


def _reshape_and_scale_image(data, max_camera_white_level, scale=False):
    """\
    plt.imshow() expects the image to have shape (size_y, size_x, num_colors) and values in [0.0 .. 1.0].
    """
    num_colors, size_y, size_x = data.shape
    plt_image = np.moveaxis(data, 0, 2)
    assert plt_image.shape == (size_y, size_x, num_colors)

    plt_image /= max_camera_white_level

    if scale:
        # brighten image so the maximum becomes 1
        plt_image /= np.nanmax(data)

        # and replace each y-column with it's maximum
        for c in range(num_colors):
            plt_image[:, :, c] = np.nanmax(plt_image[:, :, c], axis=0)

    return plt_image



import numpy as np
from pytest import approx
from scipy.ndimage.interpolation import rotate

from bayer.extraction import FastExtraction


def test_de_rotation():
    slit = 0.2 + 10 * np.exp(-0.5 * (np.arange(0, 30.0 + 1) - 15) ** 2 / 3 ** 2)
    spectrum = 0.2 + 10 * np.exp(-0.5 * (np.arange(0, 200.0 + 1) - 100) ** 2 / 40 ** 2)

    image = np.outer(slit, spectrum)

    for expected in np.arange(-180, 181, 20):
        rotated = rotate(image, angle=-expected, reshape=True, axes=(0, 1))

        fast = FastExtraction(image_layers=[rotated], sigma=3.0)
        [actual] = fast.de_rotation_angles_deg

        assert (actual == approx(expected, abs=2) or
                actual == approx(expected - 180, abs=2) or
                actual == approx(expected + 180, abs=2))



import logging
import os.path
from argparse import ArgumentParser

import matplotlib.cm
import numpy as np
from matplotlib import pyplot as plt

from bayer.extraction import FastExtraction, find_slit_in_images
from bayer.to_rgb import rawpy_to_rgb
from bayer.utils import multi_glob


def main_raw():
    import rawpy

    logging.basicConfig(level=logging.INFO)

    parser = _create_argument_parser('one or more raw files containing bayer matrices')
    args = parser.parse_args()

    for filename in multi_glob(args.filename):
        if not os.path.exists(filename):
            continue

        with rawpy.imread(filename) as raw:
            layers = rawpy_to_rgb(raw)

        fast = FastExtraction(image_layers=layers, sigma=args.sigma, clipping=args.clipping)
        _plot_file(filename, fast)


def main_fits():
    from astropy.io import fits

    parser = _create_argument_parser('one or more fits files containing images')
    args = parser.parse_args()

    for filename in multi_glob(args.filename):
        with fits.open(filename) as hdu_list:
            images = [hdu.data for hdu in hdu_list if hdu.header.get("NAXIS", 0) == 2]
            if not images:
                logging.error(f"{filename} contains no images")

            for image in images:
                extractor = FastExtraction(image_layers=[image], sigma=args.sigma)
                _plot_file(filename, extractor)


def _plot_file(filename, fast):
    zero_sigma = fast.background_mean
    three_sigma = fast.background_mean + fast.background_stddev * fast.sigma
    ten_sigma = fast.background_mean + fast.background_stddev * fast.clipping

    miny, maxy = find_slit_in_images(fast.layers, zero_sigma, scale=2)

    layers = fast.layers[:, miny:maxy, :]
    num_layers, row_count, column_count = layers.shape

    contour_levels = np.transpose([zero_sigma, three_sigma, ten_sigma])
    contour_colors = 'none green red'.split()  # none sets alpha to zero

    fs = 24
    dpi = 150
    w = column_count / dpi
    h = (fs + num_layers * (fs + row_count)) / dpi

    fig = plt.figure(figsize=(w, h), dpi=dpi)
    fig.canvas.manager.set_window_title(os.path.basename(filename))

    (layer_axes,) = fig.subplots(nrows=num_layers, ncols=1, squeeze=False).T

    for idx_layer in range(num_layers):
        layer_axis = layer_axes[idx_layer]
        layer_axis.get_xaxis().set_visible(False)
        layer_axis.get_yaxis().set_visible(False)

        # title = ((layer_titles[idx_layer] + " - ") if layer_titles else '') \
        #     + f"$\\mu_b={zero_sigma[idx_layer]:.1f}$, $\\sigma_b={fast.background_stddev[idx_layer]:.1f}$"
        # layer_axis.set_title(title)

        layer = layers[idx_layer]

        layer_axis.imshow(layer, cmap=matplotlib.cm.gray)
        layer_axis.contour(np.arange(column_count), np.arange(row_count), layer,
                           levels=contour_levels[idx_layer], colors=contour_colors)

    fig.tight_layout(pad=0.5, h_pad=0.1)
    plt.show()
    plt.close(fig)


def _create_argument_parser(filename_help):
    parser = ArgumentParser(description='Visualize segmentation of an image')
    parser.add_argument('filename', nargs='+', help=filename_help)
    parser.add_argument('--sigma', '-s', default=3.0, help='sigma used for clipping')
    parser.add_argument('--clipping', '-c', default=10.0, help='clip background at mean + clipping * stddev')
    return parser



import os
import sys
import pytest

from bayer.scripts import display_histogram, display_spectrum, visualize_segmentation


def _get_raw_filenames():

    datadir = _find('data')

    filenames = [
        'betori.CR2',
        'alpori.CR2',
        'alpcma.NEF'
    ]

    return [os.path.join(datadir, fn) for fn in filenames]


def _get_fits_filenames():

    datadir = _find('data')

    filenames = [
        'alpleo.FIT',
    ]

    return [os.path.join(datadir, fn) for fn in filenames]


def test_display_raw_histogram():

    for filename in _get_raw_filenames():
        if not os.path.exists(filename):
            continue
        sys.argv = ['dummy', filename]
        display_histogram.main_raw()


def test_display_fits_histogram():

    for filename in _get_fits_filenames():
        if not os.path.exists(filename):
            continue
        sys.argv = ['dummy', filename]
        display_histogram.main_fits()


def test_help_histogram():

    with pytest.raises(SystemExit, match='0'):
        sys.argv = ['dummy', '--help']
        display_histogram.main_raw()


def test_display_raw_spectrum():

    for filename in _get_raw_filenames():
        if not os.path.exists(filename):
            continue
        sys.argv = ['dummy', filename]
        display_spectrum.main_raw()


def test_display_fits_spectrum():

    for filename in _get_fits_filenames():
        if not os.path.exists(filename):
            continue
        sys.argv = ['dummy', filename]
        display_spectrum.main_fits()


def test_help_spectrum():

    with pytest.raises(SystemExit, match='0'):
        sys.argv = ['dummy', '--help']
        display_spectrum.main_raw()


def test_display_raw_contour():

    for filename in _get_raw_filenames():
        if not os.path.exists(filename):
            continue
        sys.argv = ['dummy', filename]
        visualize_segmentation.main_raw()


def test_display_fits_contour():

    for filename in _get_fits_filenames():
        if not os.path.exists(filename):
            continue
        sys.argv = ['dummy', filename]
        visualize_segmentation.main_fits()


def test_help_contour():

    with pytest.raises(SystemExit, match='0'):
        sys.argv = ['dummy', '--help']
        visualize_segmentation.main_raw()


def _find(target):

    here = os.path.dirname(__file__)

    while True:
        join = os.path.join(here, target)
        parent, _ = os.path.split(here)

        if os.path.exists(join):
            return join

        elif parent and parent != here:
            here = parent

        else:
            return None
